// ..................................................................
//
//                             mdTray
//
//              Copyright  1997 by Martin Djerns
//
// ..................................................................
// $Id: mdTray.pas 1.3 1998/08/19 20:24:02 md Exp md $
// ..................................................................
// Initial Date : 2 November 1997 - MD
// 12 November 1997 - Version 1.0 released
// 19 November 1997 -
// + Check for avalibility of a "real" tray class
//   (coursed an error durring design time!)
// - SetHandle and GetHandle for TmdTrayIcon, since a TmdTrayIcon
//   is not actepted by the default Icon Property editor, and
//   the Handle property can not be overwritten...
// 19 November 1997 - Version 1.1 released
// 6  February 1998 -
// % Simpel should be Simple (Sorry) ;-)
// 10th April 1998 - Version 1.2 released
// 5th Juni 1998
// + Added support for two trays in one application
// 19th August 1998 - Version 1.3 released
// ..................................................................
// Description :
// - I were fed up with the TrayIcon components which I got from
//   the internet. They all missed something, like being able to
//   use icons from the resources, or updating the tray when a new
//   icon is added.
// - Uses a own Icon, for being able to being told when the icon
//   is changed so the tray can be updated.
// - Is tried in Delphi 2.0 and 3.0.
// - When wanting to change the icon via a handle change, then use
//   mdTray.SetIconHandle(AHandle) instead of
//   mdTray.Icon.Handle := AHandle, or call the mdTray.IconChanged
//   manually-
// ..................................................................
// Known issues :
// - Rightclick in tray icon do not always show popup menu
// - Based on mdMsgComponent (= meight requere a bit resources)
// - Meight not work without a form!
// - Loaded do not work! (Active = false is not done, and
//   resources equal nothing is also ignored!!! (Eric Sellers)
// ..................................................................

unit mdTray;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  mdMsgCmp, ShellAPI, Menus;

Const
  WM_TrayMessage = WM_User + 9100; // Unique ????
  defTrayActive = True;

type
  TTipString = String[63]; // Since a tip maximum can be 63 chars long

  TWMTrayMessage = Record  // Windows message from the tray...
    Msg: Cardinal;
    Id,
    Command,
    Result: Longint;
  end;

  TmdTray = class(TmdMsgComponent)
  private
    { Private declarations }
    FActive : Boolean;
    FTrayData : TNotifyIconData;
    FIconResourceNames : TStrings;
    FIconIndex : Integer;
    FIcon : TIcon;
    FPopupMenu : TPopupMenu;

    FTipChanged : Boolean;
    FTrayId : Integer;

    FShowMenu : TNotifyEvent;
    FSimpleFunction : TNotifyEvent;
    FDefaultFunction : TNotifyEvent;

    Procedure SetActive(Value : Boolean);
    Function GetActive : Boolean;
    Procedure SetIcon(Value : TIcon);
    Procedure SetIconResourceNames(Value : TStrings);
    Procedure SetIconIndex(Value : Integer);
    Procedure SetTip(Value : TTipString);
    Function GetTip : TTipString;
    Procedure SetPopupMenu(Value : TPopupMenu);
  protected
    { Protected declarations }
    Procedure SetName(const NewName: TComponentName); Override;

    Procedure WMTrayMessage(Var TrayMsg : TWMTrayMessage); Message WM_TrayMessage;
    Procedure CreateTrayIcon; Virtual;
    Procedure FreeTrayIcon; Virtual;
    Procedure UpdateTrayIcon; Virtual;
    Procedure Popup(Pt : TPoint); Virtual;

    Procedure DoShowMenu;
    Procedure DoSimpleFunction;
    Procedure DoDefaultFunction;
  public
    { Public declarations }
    Constructor Create(AOwner : TComponent); Override;
    Destructor Destroy; Override;
    Procedure Loaded; Override;
    Procedure Notification(AComponent : TComponent;
                           Operation : TOperation); Override;
    Function GetIconHandle : HICON;

    Procedure IconChanged; Virtual;

    Procedure SetIconHandle(Value : HICON);
    Procedure First;
    Procedure Next;
    Procedure Prev;
    Procedure Last;
  published
    // Properies
    Property Active : Boolean
      Read GetActive Write SetActive Default defTrayActive;
    Property Icon : TIcon
      Read FIcon Write SetIcon;
    Property IconResources : TStrings
      Read FIconResourceNames Write SetIconResourceNames;
    Property IconIndex : Integer
      Read FIconIndex Write SetIconIndex;
    Property PopupMenu : TPopupMenu
      Read FPopupMenu Write SetPopupMenu;
    Property Tip : TTipString
      Read GetTip Write SetTip;

    // Events
    Property ShowMenu : TNotifyEvent
      Read FShowMenu Write FShowMenu;
    Property SimpleFunction : TNotifyEvent
      Read FSimpleFunction Write FSimpleFunction;
    Property DefaultFunction : TNotifyEvent
      Read FDefaultFunction Write FDefaultFunction;
  end;

procedure Register;

implementation

Const
  mdTrayId : Integer = 0;

// ------ Tray Icon function ------

Type
// This icon class should make sure that
// we always get acknowledged when the icon
// is changed, also when only the handle is
// canged.
  TmdTrayIcon = class(TIcon)
  Private
    FTray : TmdTray;
  Public
    procedure Assign(Source: TPersistent); Override;
  end;

Procedure TmdTrayIcon.Assign(Source: TPersistent);
Begin
  Inherited Assign(Source);
  If FTray <> NIL Then
    FTray.IconChanged;   // Acknowlegde TmdTray about new icon
end;

// ------ TmdTrayStrings     -------

Type
  TmdTrayResources = class(TStringList)
  Private
    FTray : TmdTray;
  Protected
    Procedure Put(Index: Integer; Const S: String); Override;
  Public
    Constructor Create(Tray : TmdTray); Virtual;
  end;

Constructor TmdTrayResources.Create(Tray : TmdTray);
Begin
  Inherited Create;
  FTray := Tray;  // We need to know what thay to acknowledge
end;

Procedure TmdTrayResources.Put(Index : Integer; Const S : String);
Begin
  Inherited Put(Index,S);
  If FTray <> NIL Then
    FTray.IconChanged;  // Can be changed...
end;

// ------ Standard functions -------

Constructor TmdTray.Create(AOwner : TComponent);
Begin
  Inherited Create(AOwner);
  FActive := defTrayActive;
  FIcon := TmdTrayIcon.Create;
  TmdTrayIcon(FIcon).FTray := Self;
  FIconResourceNames := TmdTrayResources.Create(Self);
  FIconResourceNames.Add('MAINICON');  // Use the project icon as default
end;

Destructor TmdTray.Destroy;
Begin
  FreeTrayIcon;
  TmdTrayIcon(FIcon).Free;
  FIconResourceNames.Free;
  Inherited Destroy;
end;

Procedure TmdTray.Loaded;
Begin
  Inherited Loaded;
  CreateTrayIcon;
end;

Procedure TmdTray.Notification(AComponent : TComponent;
                               Operation : TOperation);
Begin
  Inherited Notification(AComponent, Operation);
  If (Operation = opRemove) AND (AComponent = FPopupMenu) Then
    PopupMenu := NIL;
end;

Procedure TmdTray.SetName(const NewName: TComponentName);
Begin
  Inherited SetName(NewName);
  If NOT FTipChanged Then
  Begin
    Tip := NewName;
    FTipChanged := False;  // Still not changed (by user)
  end;
end;

// ------ Property functions  ------

Procedure TmdTray.SetActive(Value : Boolean);
Begin
  If Active = Value Then
    Exit;
  FActive := Value;
  If FActive Then
    CreateTrayIcon
  else
    FreeTrayIcon;
end;

Function TmdTray.GetActive : Boolean;
Begin
  // If the component is edited in the Delphi IDE
  // return the wanted value, and not the
  // real state
  If (NOT (csDesigning IN ComponentState)) AND
     (FTrayData.hIcon = 0) Then
    Result := False
  else
    Result := FActive;
end;

Procedure TmdTray.SetIcon(Value : TIcon);
Begin
  FIcon.Assign(Value);
  UpdateTrayIcon;
end;

Procedure TmdTray.SetIconResourceNames(Value : TStrings);
Begin
  FIconResourceNames.Assign(Value);
  UpdateTrayIcon;
end;

Procedure TmdTray.SetIconIndex(Value : Integer);
Begin
  If Value = FIconIndex Then
    Exit;
  FIconIndex := Value;
  If FIcon.Handle = 0 Then
    UpdateTrayIcon;
end;

Procedure TmdTray.SetTip(Value : TTipString);
Var
  V: AnsiString;
Begin
  // Make ourself aware of that the user have changed the tip value
  FTipChanged := True;
  V := Value;
  // Store the tip directly in the data structure
  StrCopy(FTrayData.szTip, PChar(V));
  UpdateTrayIcon;
end;

Function TmdTray.GetTip : TTipString;
Begin
  Result := TTipString(FTrayData.szTip);
end;

Procedure TmdTray.SetPopupMenu(Value : TPopupMenu);
Begin
  If Value <> FPopupMenu Then
  Begin
    FPopupMenu := Value;
    // So we are notified if the popup menu is released
    If Value <> NIL Then
      Value.FreeNotification(Self);
  end;
end;

Procedure TmdTray.Popup(Pt : TPoint);
Begin
  If csDesigning IN ComponentState Then
    Exit;
  If (PopupMenu <> NIL) AND (PopupMenu.AutoPopUp) Then
  Begin
    // So we get the shortcuts, and the menu react to lost focus...
    SetForegroundWindow(Handle);
    PopupMenu.PopupComponent := Self;
    PopupMenu.Popup(Pt.X,Pt.Y);
//    PostMessage(0,0,0,0);  // Some people say that this line should
                             // help getting the focus faster, but I
                             // can not see the difference !
  end;
end;

// ------ Public functions    ------

// "Internal" function to get the handle of the
// icon which should be used.
// The Icon property have first priority,
// and if that one have Handle which is
// zero, then load one from the resource.
Function TmdTray.GetIconHandle : HICON;
Begin
  Result := FIcon.Handle;
  If (Result = 0) AND
     (FIconResourceNames <> NIL) AND
     (FIconResourceNames.Count > 0) AND
     (FIconResourceNames.Count > FIconIndex) Then
    Result := LoadIcon(HInstance, PChar(FIconResourceNames[FIconIndex]));
end;

Procedure TmdTray.SetIconHandle(Value : HICON);
Begin
  FIcon.Handle := Value;
  IconChanged;
end;

Procedure TmdTray.IconChanged;
Begin
  // Only update if there is something to update
  If FTrayData.hIcon <> GetIconHandle Then
    UpdateTrayIcon;
end;

Procedure TmdTray.First;
Begin
  IconIndex := 0;
end;

Procedure TmdTray.Prev;
Begin
  If FIconResourceNames.Count = 0 Then
    Exit;
  If FIconIndex = 0 Then
    IconIndex := FIconResourceNames.Count-1
  else
    IconIndex := FIconIndex - 1;
end;

Procedure TmdTray.Next;
Begin
  If FIconIndex >= FIconResourceNames.Count-1 Then
    IconIndex := 0
  else
    IconIndex := FIconIndex + 1;
end;

Procedure TmdTray.Last;
Begin
  IconIndex := FIconResourceNames.Count - 1;
end;

// ------ Tray Icon functions ------

Procedure TmdTray.CreateTrayIcon;
Begin
  If csDesigning IN ComponentState Then
    Exit;
  With FTrayData do
  Begin
    cbSize := SizeOf(TNotifyIconData);
    If Handle = 0 Then
      Handle := ParentForm.Handle;
    Wnd := Handle;
    If FTrayId = 0 Then
    Begin
      Inc(mdTrayId);
      FTrayId := mdTrayId;
    end;
    uID := FTrayId;
    uCallBackMessage := WM_TrayMessage; // The value of the message
                                        // which we want to receive
    hIcon := GetIconHandle;
    uFlags := NIF_MESSAGE or NIF_ICON;
    If Tip <> '' Then  // Only use tip if a tip string is set
      FTrayData.uFlags := FTrayData.uFlags or NIF_TIP
    else
      FTrayData.uFlags := FTrayData.uFlags AND (NOT NIF_TIP);
  end;
  // Only create the icon if there is set a icon
  If (FTrayData.hIcon <> 0) AND (FTrayData.Wnd <> 0) Then
    Shell_NotifyIcon (NIM_ADD, @FTrayData);
end;

Procedure TmdTray.FreeTrayIcon;
Begin
  If csDesigning IN ComponentState Then
    Exit;
  FTrayData.uFlags := 0;
  Shell_NotifyIcon (NIM_DELETE, @FTrayData);
end;

Procedure TmdTray.UpdateTrayIcon;
Begin
  If csDesigning IN ComponentState Then
    Exit;
  If NOT Active Then
    Exit;
  If Tip <> '' Then
    FTrayData.uFlags := FTrayData.uFlags or NIF_TIP
  else
    FTrayData.uFlags := FTrayData.uFlags AND (NOT NIF_TIP);

  FTrayData.hIcon := GetIconHandle;
  Shell_NotifyIcon (NIM_MODIFY, @FTrayData);
end;

Procedure TmdTray.WMTrayMessage(Var TrayMsg : TWMTrayMessage);
Begin
  TrayMsg.Result := LongInt(False);
  If TrayMsg.Id = FTrayId Then
  Begin
    Case TrayMsg.Command of
      WM_RButtonDown     :
        DoShowMenu;
      WM_LButtonDown     :
        DoSimpleFunction;
      WM_LButtonDblClk   :
        DoDefaultFunction;
    else
      Exit;
    end;
    TrayMsg.Result := LongInt(True);
  end;
end;

// ------ Event functions     ------

Procedure TmdTray.DoShowMenu;
Var
  Pt : TPoint;
Begin
  If Assigned(FShowMenu) Then
    FShowMenu(Self);
  GetCursorPos(Pt);    // Get the position to popup at
  Popup(Pt);
end;

Procedure TmdTray.DoSimpleFunction;
Begin
  If Assigned(FSimpleFunction) Then
    FSimpleFunction(Self);
end;

Procedure TmdTray.DoDefaultFunction;
Begin
  If Assigned(FDefaultFunction) Then
    FDefaultFunction(Self);
end;

// ------ VCL funstions       ------

procedure Register;
begin
  RegisterComponents('mdVCL', [TmdTray]);
end;

end.
